"""
Vendored modules - see licenses in /licenses.

Currently consists just of nptyping, as it is no longer maintained
and pins a version of numpy<2, and we have to do an increasing
number of awkward monkeypatches for perf and customization reasons.

This vendored module will slowly be worked out of the code and
all its functionality replaced.
"""
